@@ -1,23 +1,16 @@
-# -*- coding: utf8 -*-
-
 import os
+import string
+import random
 import logging
 from typing import List, Union
 
 import requests
 
-
 logger = logging.getLogger(__name__)
 
-
+ACCESS_TOKEN = None
 ACCESS_TOKEN_ENV_NAME = 'SLACK_ACCESS_TOKEN'
 
-SLACK_API_URL = 'https://slack.com/api'
-ACCESS_TOKEN = os.getenv(ACCESS_TOKEN_ENV_NAME)
-
-DEFAULT_RECORDS_LIMIT = 100
-DEFAULT_REQUEST_TIMEOUT = 180
-
 COLOR_MAP = {
     'green': '#008000',
     'gray': '#808080',
@@ -43,6 +36,11 @@ COLOR_MAP = {
 }
 
 
+def _random_string(length):
+    letters = string.ascii_lowercase
+    return ''.join(random.choice(letters) for i in range(length))
+
+
 class SlackError(requests.exceptions.RequestException):
     pass
 
@@ -61,7 +59,7 @@ class DictConvertibleObject:
 
     def to_dict(self):
         raise NotImplementedError(
-            f'Object "{self.__class__.__name__}" does not implemented "to_dict" method',
+            'Object "{}" does not implemented "to_dict" method'.format(self.__class__.__name__),
         )
 
 
@@ -90,7 +88,6 @@ class AttachmentField(DictConvertibleObject):
 
 
 class Attachment(DictConvertibleObject):
-
     Field = AttachmentField
 
     def __init__(self, *,
@@ -189,12 +186,13 @@ class Attachment(DictConvertibleObject):
 
         if self.fields:
             data['fields'] = [f.to_dict() for f in self.fields]
+            if self.mrkdwn:
+                data['mrkdwn_in'].append('fields')
 
         return data
 
 
 class BaseBlock(DictConvertibleObject):
-
     __type__ = None
 
     def __init__(self, *, mrkdwn: bool = True, block_id: str = None):
@@ -216,7 +214,6 @@ class BaseBlock(DictConvertibleObject):
 
 
 class BaseBlockField(DictConvertibleObject):
-
     __type__ = None
 
     def __init__(self, *, mrkdwn=True):
@@ -234,6 +231,26 @@ class BaseBlockField(DictConvertibleObject):
         return {}
 
 
+class HeaderBlock(BaseBlock):
+    __type__ = 'header'
+
+    def __init__(self, text: str, **kwargs):
+        kwargs['mrkdwn'] = False
+        super().__init__(**kwargs)
+
+        self.text = text
+
+    def to_dict(self):
+        data = super().to_dict()
+
+        data['text'] = {
+            'type': self.content_type,
+            'text': self.text,
+        }
+
+        return data
+
+
 class SimpleTextBlockField(BaseBlockField):
 
     def __init__(self, text: str, *, emoji: bool = None, **kwargs):
@@ -255,7 +272,6 @@ class SimpleTextBlockField(BaseBlockField):
 
 
 class SimpleTextBlock(BaseBlock):
-
     __type__ = 'section'
 
     Field = SimpleTextBlockField
@@ -281,12 +297,10 @@ class SimpleTextBlock(BaseBlock):
 
 
 class DividerBlock(BaseBlock):
-
     __type__ = 'divider'
 
 
 class ImageBlock(BaseBlock):
-
     __type__ = 'image'
 
     def __init__(self, image_url, *, title: str = None, alt_text: str = None, **kwargs):
@@ -331,7 +345,6 @@ class ContextBlockTextElement(BaseBlockField):
 
 
 class ContextBlockImageElement(BaseBlockField):
-
     __type__ = 'image'
 
     def __init__(self, image_url, alt_text: str = None):
@@ -352,7 +365,6 @@ class ContextBlockImageElement(BaseBlockField):
 
 
 class ContextBlock(BaseBlock):
-
     __type__ = 'context'
 
     TextElement = ContextBlockTextElement
@@ -375,89 +387,258 @@ def init_color(name, code):
     COLOR_MAP[name] = code
 
 
-def call_resource(resource: Resource, *, raise_exc: bool = False, **kwargs):
-    assert ACCESS_TOKEN is not None, f'Please export "{ACCESS_TOKEN_ENV_NAME}" environment variable'
+class Message:
 
-    kwargs.setdefault('timeout', DEFAULT_REQUEST_TIMEOUT)
+    def __init__(self, client, response,
+                 text: str = None,
+                 raise_exc=False,
+                 attachments: List[Attachment] = None,
+                 blocks: List[BaseBlock] = None):
+        self._client = client
+        self._response = response
+        self._raise_exc = raise_exc
 
-    url = f'{SLACK_API_URL}/{resource.handle}'
+        self.text = text
+        self.attachments = attachments or []
+        self.blocks = blocks or []
 
-    kwargs['headers'] = {
-        'Authorization': f'Bearer {ACCESS_TOKEN}',
-        'Content-Type': 'application/json; charset=utf-8',
-    }
+        self.__lock_thread = False
 
-    response = requests.request(resource.method, url, **kwargs)
+    @property
+    def response(self):
+        return self._response
 
-    logger.debug(response.content)
+    def _lock_thread(self):
+        self.__lock_thread = True
 
-    if raise_exc:
-        response.raise_for_status()
+    def add_reaction(self, name, raise_exc=False):
+        json = self._response.json()
+        data = {
+            'name': name,
+            'channel': json['channel'],
+            'timestamp': json['message']['ts'],
+        }
+        return self._client.call_resource(
+            Resource('reactions.add', 'POST'),
+            json=data, raise_exc=raise_exc,
+        )
 
-        json = response.json()
+    def remove_reaction(self, name, raise_exc=False):
+        json = self._response.json()
+        data = {
+            'name': name,
+            'channel': json['channel'],
+            'timestamp': json['message']['ts'],
+        }
+        return self._client.call_resource(
+            Resource('reactions.remove', 'POST'),
+            json=data, raise_exc=raise_exc,
+        )
 
-        if not json['ok']:
-            logger.error(response.content)
-            raise SlackError(response=response)
+    def send_to_thread(self, **kwargs):
+        if self.__lock_thread:
+            raise SlackError('Cannot open thread for thread message')
 
-    return response
+        json = self._response.json()
+        thread_ts = json['message']['ts']
+        kwargs.update(thread_ts=thread_ts)
+
+        message = self._client.send_notify(json['channel'], **kwargs)
+
+        lock_thread = getattr(message, '_lock_thread')
+        lock_thread()
+
+        return message
+
+    def update(self):
+        json = self._response.json()
+        data = {
+            'channel': json['channel'],
+            'ts': json['message']['ts'],
+        }
+
+        if self.text:
+            data['text'] = self.text
+
+        if self.blocks:
+            data['blocks'] = [b.to_dict() for b in self.blocks]
+
+        if self.attachments:
+            data['attachments'] = [a.to_dict() for a in self.attachments]
+
+        return self._client.call_resource(
+            Resource('chat.update', 'POST'),
+            json=data, raise_exc=self._raise_exc,
+        )
+
+    def delete(self):
+        json = self._response.json()
+        data = {
+            'channel': json['channel'],
+            'ts': json['message']['ts'],
+        }
+        return self._client.call_resource(
+            Resource('chat.update', 'POST'),
+            json=data, raise_exc=self._raise_exc,
+        )
+
+    def upload_file(self, file, **kwargs):
+        json = self._response.json()
+        kwargs.update(thread_ts=json['message']['ts'])
+        return self._client.upload_file(json['channel'], file, **kwargs)
 
 
-def resource_iterator(resource: Resource, from_key: str, *,
-                      cursor: str = None,
-                      raise_exc: bool = False,
-                      limit: int = DEFAULT_RECORDS_LIMIT):
-    params = {'limit': limit}
+class Slack(requests.Session):
+    API_URL = 'https://slack.com/api'
 
-    if cursor:
-        params['cursor'] = cursor
+    DEFAULT_RECORDS_LIMIT = 100
+    DEFAULT_REQUEST_TIMEOUT = 180
 
-    response = call_resource(resource, params=params, raise_exc=raise_exc)
-    data = response.json()
+    def __init__(self, token):
+        super(Slack, self).__init__()
 
-    for item in data[from_key]:
-        yield item
+        self.headers['Authorization'] = 'Bearer {}'.format(token)
+        self.headers['Content-Type'] = 'application/json; charset=utf-8'
 
-    cursor = data.get('response_metadata', {}).get('next_cursor')
+    @classmethod
+    def from_env(cls):
+        token = ACCESS_TOKEN or os.getenv(ACCESS_TOKEN_ENV_NAME)
+        assert token is not None, 'Please export "{}" environment variable'.format(ACCESS_TOKEN_ENV_NAME)
+        return cls(token)
 
-    if cursor:
-        yield from resource_iterator(resource, from_key, limit=limit, cursor=cursor, raise_exc=raise_exc)
+    def call_resource(self, resource: Resource, *, raise_exc: bool = False, **kwargs):
+        kwargs.setdefault('timeout', self.DEFAULT_REQUEST_TIMEOUT)
+
+        url = '{}/{}'.format(self.API_URL, resource.handle)
+        response = self.request(resource.method, url, **kwargs)
+
+        logger.debug(response.content)
+
+        if raise_exc:
+            response.raise_for_status()
+
+            json = response.json()
+
+            if not json['ok']:
+                logger.error(response.content)
+                raise SlackError(response=response)
+
+        return response
+
+    def resource_iterator(self,
+                          resource: Resource, from_key: str, *,
+                          cursor: str = None,
+                          raise_exc: bool = False,
+                          limit: int = None):
+        params = {'limit': limit}
+
+        if cursor:
+            params['cursor'] = cursor
+
+        response = self.call_resource(resource, params=params, raise_exc=raise_exc)
+        data = response.json()
+
+        for item in data[from_key]:
+            yield item
+
+        cursor = data.get('response_metadata', {}).get('next_cursor')
+
+        if cursor:
+            yield from self.resource_iterator(
+                resource, from_key,
+                limit=limit or self.DEFAULT_RECORDS_LIMIT, cursor=cursor, raise_exc=raise_exc,
+            )
+
+    def upload_file(self,
+                    channel, file, *,
+                    title: str = None,
+                    content: str = None,
+                    filename: str = None,
+                    thread_ts: str = None,
+                    filetype: str = 'text',
+                    raise_exc: bool = False):
+        data = {
+            'channels': channel,
+            'filetype': filetype,
+        }
+        if isinstance(file, str) and content:
+            filename = file
+            data.update(content=content, filename=filename)
+        elif isinstance(file, str) and not content:
+            data.update(filename=os.path.basename(file))
+            with open(file, 'r') as f:
+                data.update(content=f.read())
+        else:
+            data.update(content=file.read(), filename=filename or _random_string(7))
+
+        if title:
+            data.update(title=title)
+        if thread_ts:
+            data.update(thread_ts=thread_ts)
+
+        return self.call_resource(
+            Resource('files.upload', 'POST'),
+            data=data,
+            raise_exc=raise_exc,
+            headers={
+                'Content-Type': 'application/x-www-form-urlencoded',
+            },
+        )
+
+    def send_notify(self,
+                    channel, *,
+                    text: str = None,
+                    username: str = None,
+                    icon_url: str = None,
+                    icon_emoji: str = None,
+                    link_names: bool = True,
+                    raise_exc: bool = False,
+                    attachments: List[Attachment] = None,
+                    blocks: List[BaseBlock] = None,
+                    thread_ts: str = None):
+        data = {
+            'channel': channel,
+            'link_names': link_names,
+        }
+
+        if username:
+            data['as_user'] = False
+            data['username'] = username
+        else:
+            data['as_user'] = True
+
+        if text:
+            data['mrkdwn'] = True
+            data['text'] = text
+
+        if icon_url and not data['as_user']:
+            data['icon_url'] = icon_url
+
+        if icon_emoji and not data['as_user']:
+            data['icon_emoji'] = icon_emoji
+
+        if blocks:
+            data['blocks'] = [b.to_dict() for b in blocks]
+
+        if attachments:
+            data['attachments'] = [a.to_dict() for a in attachments]
+
+        if thread_ts:
+            data['thread_ts'] = thread_ts
+
+        response = self.call_resource(
+            Resource('chat.postMessage', 'POST'), raise_exc=raise_exc, json=data,
+        )
+        return Message(self, response, text=text, raise_exc=raise_exc, blocks=blocks, attachments=attachments)
 
 
-def send_notify(channel, *,
-                text: str = None,
-                username: str = None,
-                icon_url: str = None,
-                icon_emoji: str = None,
-                link_names: bool = True,
-                raise_exc: bool = False,
-                attachments: List[Attachment] = None,
-                blocks: List[BaseBlock] = None):
-    data = {
-        'channel': channel,
-        'link_names': link_names,
-    }
+def call_resource(*args, **kwargs):
+    return Slack.from_env().call_resource(*args, **kwargs)
 
-    if username:
-        data['as_user'] = False
-        data['username'] = username
-    else:
-        data['as_user'] = True
 
-    if text:
-        data['mrkdwn'] = True
-        data['text'] = text
+def resource_iterator(*args, **kwargs):
+    return Slack.from_env().resource_iterator(*args, **kwargs)
 
-    if icon_url and not data['as_user']:
-        data['icon_url'] = icon_url
 
-    if icon_emoji and not data['as_user']:
-        data['icon_emoji'] = icon_emoji
-
-    if blocks:
-        data['blocks'] = [b.to_dict() for b in blocks]
-
-    if attachments:
-        data['attachments'] = [a.to_dict() for a in attachments]
-
-    return call_resource(Resource('chat.postMessage', 'POST'), raise_exc=raise_exc, json=data)
+def send_notify(*args, **kwargs):
+    return Slack.from_env().send_notify(*args, **kwargs)
